"""FINRA OTC Aggregate Model."""

from typing import Any, Dict, List, Literal, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.otc_aggregate import (
    OTCAggregateData,
    OTCAggregateQueryParams,
)
from openbb_finra.utils.helpers import get_full_data
from pydantic import Field


class FinraOTCAggregateQueryParams(OTCAggregateQueryParams):
    """FINRA OTC Aggregate Query."""

    tier: Literal["T1", "T2", "OTCE"] = Field(
        default="T1",
        description=""""T1 - Securities included in the S&P 500, Russell 1000 and selected exchange-traded products;
        T2 - All other NMS stocks; OTC - Over-the-Counter equity securities""",
    )
    is_ats: bool = Field(
        default=True, description="ATS data if true, NON-ATS otherwise"
    )


class FinraOTCAggregateData(OTCAggregateData):
    """FINRA OTC Aggregate Data."""

    __alias_dict__ = {
        "share_quantity": "totalWeeklyShareQuantity",
        "trade_quantity": "totalWeeklyTradeCount",
        "update_date": "lastUpdateDate",
    }


class FinraOTCAggregateFetcher(
    Fetcher[FinraOTCAggregateQueryParams, List[FinraOTCAggregateData]]
):
    """Transform the query, extract and transform the data from the FINRA endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FinraOTCAggregateQueryParams:
        """Transform query params."""
        return FinraOTCAggregateQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FinraOTCAggregateQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Extract the data from the FINRA endpoint."""
        return get_full_data(query.symbol, query.tier, query.is_ats)

    @staticmethod
    def transform_data(
        query: FinraOTCAggregateQueryParams, data: List[Dict], **kwargs: Any
    ) -> List[FinraOTCAggregateData]:
        """Transform the data."""
        return [FinraOTCAggregateData.model_validate(d) for d in data if d]



"""Utility for FINRA data storage.

This was created as a way to handle short interest data from the FINRA.
The files do not change, so there is no need to download them every time.
"""
import random
import sqlite3
from io import StringIO
from pathlib import Path
from typing import List

import requests
from openbb_core.app.utils import get_user_cache_directory
from openbb_finra.utils.helpers import get_short_interest_dates
from pandas import read_csv

DB_PATH = Path(get_user_cache_directory()) / "caches/finra_short_volume.db"
DB_PATH.parent.mkdir(parents=True, exist_ok=True)


def get_cached_dates() -> List:
    """Return the dates that are cached in the DB file."""
    cnx = sqlite3.connect(DB_PATH)
    cursor = cnx.cursor()

    # Check if the table exists
    cursor.execute(
        "SELECT * FROM sqlite_master WHERE type='table' AND name='short_interest'"
    )
    if cursor.fetchone() is None:
        # The table does not exist
        return []

    # If the table exists, fetch the data
    cursor.execute("SELECT distinct settlementDate FROM short_interest")
    result = cursor.fetchall()
    return [row[0] for row in result]


def get_data_from_date_and_store(date):
    """Get data from a specific date and place it in the cache."""
    url = f"https://cdn.finra.org/equity/otcmarket/biweekly/shrt{date}.csv"
    # add a random string to user agent to avoid getting blocked
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
        + str(random.randint(0, 9))  # noqa: S311
    }
    req = requests.get(url, headers=headers, timeout=1)
    if req.status_code != 200:
        return
    data = read_csv(StringIO(req.text), delimiter="|")
    data = data.drop(
        columns=[
            "accountingYearMonthNumber",
            "issuerServicesGroupExchangeCode",
            "stockSplitFlag",
            "revisionFlag",
        ]
    )
    data.to_sql("short_interest", sqlite3.connect(DB_PATH), if_exists="append")


def prepare_data():
    """Prepare the data."""
    date_list = get_short_interest_dates()
    cached_urls = get_cached_dates()
    for date in date_list:
        if f"{date[:4]}-{date[4:6]}-{date[6:]}" not in cached_urls:
            get_data_from_date_and_store(date)



# Helper functions for FINRA API
import datetime
from typing import List

import requests


def get_finra_weeks(tier: str = "T1", is_ats: bool = True):
    """Fetches the available weeks from FINRA that can be used."""
    request_header = {"Accept": "application/json", "Content-Type": "application/json"}

    request_data = {
        "compareFilters": [
            {
                "compareType": "EQUAL",
                "fieldName": "summaryTypeCode",
                "fieldValue": "ATS_W_SMBL" if is_ats else "OTC_W_SMBL",
            },
            {
                "compareType": "EQUAL",
                "fieldName": "tierIdentifier",
                "fieldValue": tier,
            },
        ],
        "delimiter": "|",
        "fields": ["weekStartDate"],
        "limit": 52,
        "quoteValues": False,
        "sortFields": ["-weekStartDate"],
    }

    response = requests.post(
        "https://api.finra.org/data/group/otcMarket/name/weeklyDownloadDetails",
        headers=request_header,
        json=request_data,
        timeout=3,
    )

    return response.json() if response.status_code == 200 else []


def get_finra_data(symbol, week_start, tier: str = "T1", is_ats: bool = True):
    req_hdr = {
        "Accept": "application/json",
        "Content-Type": "application/json",
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) ",
    }

    filters = [
        {
            "compareType": "EQUAL",
            "fieldName": "weekStartDate",
            "fieldValue": week_start,
        },
        {"compareType": "EQUAL", "fieldName": "tierIdentifier", "fieldValue": tier},
        {
            "compareType": "EQUAL",
            "description": "",
            "fieldName": "summaryTypeCode",
            "fieldValue": "ATS_W_SMBL" if is_ats else "OTC_W_SMBL",
        },
    ]

    if symbol:
        filters.append(
            {
                "compareType": "EQUAL",
                "fieldName": "issueSymbolIdentifier",
                "fieldValue": symbol,
            }
        )

    req_data = {
        "compareFilters": filters,
        "delimiter": "|",
        "fields": [
            "issueSymbolIdentifier",
            "totalWeeklyShareQuantity",
            "totalWeeklyTradeCount",
            "lastUpdateDate",
        ],
        "limit": 5000,
        "quoteValues": False,
        "sortFields": ["totalWeeklyShareQuantity"],
    }
    response = requests.post(
        "https://api.finra.org/data/group/otcMarket/name/weeklySummary",
        headers=req_hdr,
        json=req_data,
        timeout=2,
    )
    return response


def get_full_data(symbol, tier: str = "T1", is_ats: bool = True):
    weeks = [week["weekStartDate"] for week in get_finra_weeks(tier, is_ats)]

    data = []
    for week in weeks:
        response = get_finra_data(symbol, week, tier, is_ats)
        r_json = response.json()
        if response.status_code == 200 and r_json:
            data.append(response.json()[0])

    return data


def get_adjusted_date(year, month, day):
    # Get the date
    date = datetime.date(year, month, day)

    # If the date is a Saturday, subtract one day
    if date.weekday() == 5:
        date -= datetime.timedelta(days=1)
    # If the date is a Sunday, subtract two days
    elif date.weekday() == 6:
        date -= datetime.timedelta(days=2)

    return date


def get_short_interest_dates() -> List[str]:
    """
    Get a list of dates for which the short interest data is available.  It is reported on the 15th and the
    last day of each month,but if the date falls on a weekend, the date is adjusted to the closest friday.
    """

    def get_adjusted_date(year, month, day):
        """If the date falls on a weekend, find the closest date"""
        # Get the date
        date = datetime.date(year, month, day)

        # If the date is a Saturday, subtract one day
        if date.weekday() == 5:
            date -= datetime.timedelta(days=1)
        # If the date is a Sunday, subtract two days
        elif date.weekday() == 6:
            date -= datetime.timedelta(days=2)

        return date

    start_year = 2021
    today = datetime.date.today()  # Get today's date
    end_year = today.year
    dates_list = []

    for yr in range(start_year, end_year + 1):
        start_month = 7 if yr == start_year else 1
        end_month = 12 if yr < today.year else today.month - 1
        for month in range(start_month, end_month + 1):  # Start from July for 2021
            # Date for the 15th of the month
            date_15 = get_adjusted_date(yr, month, 15)
            dates_list.append(date_15.strftime("%Y%m%d"))

            # Date for the last day of the month
            if month == 2:  # February
                last_day = (
                    29 if (yr % 4 == 0 and yr % 100 != 0) or (yr % 400 == 0) else 28
                )
            elif month in [4, 6, 9, 11]:  # Months with 30 days
                last_day = 30
            else:  # Months with 31 days
                last_day = 31

            last_date = get_adjusted_date(yr, month, last_day)
            dates_list.append(last_date.strftime("%Y%m%d"))

    # Manually replace '20220415' with '20220414' due to holiday
    if "20220415" in dates_list:
        index = dates_list.index("20220415")
        dates_list[index] = "20220414"

    return dates_list



"""FINRA Equity Short Interest Model."""

import sqlite3
from typing import Any, Dict, List, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_short_interest import (
    ShortInterestData,
    ShortInterestQueryParams,
)
from openbb_finra.utils.data_storage import DB_PATH, prepare_data


class FinraShortInterestQueryParams(ShortInterestQueryParams):
    """FINRA Equity Short Interest Query."""


class FinraShortInterestData(ShortInterestData):
    """FINRA Equity Short Interest Data."""

    __alias_dict__ = {
        "symbol": "symbolCode",
        "issue_name": "issueName",
        "market_class": "marketClassCode",
        "current_short_position": "currentShortPositionQuantity",
        "previous_short_position": "previousShortPositionQuantity",
        "avg_daily_volume": "averageDailyVolumeQuantity",
        "days_to_cover": "daysToCoverQuantity",
        "change": "changePreviousNumber",
        "change_pct": "changePercent",
        "settlement_date": "settlementDate",
    }


class FinraShortInterestFetcher(
    Fetcher[FinraShortInterestQueryParams, List[FinraShortInterestData]]
):
    """Transform the query, extract and transform the data from the FINRA endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FinraShortInterestQueryParams:
        """Transform query params."""
        return FinraShortInterestQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FinraShortInterestQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Extract the data from the Finra endpoint."""
        # Put the data in the cache
        prepare_data()
        # Get the data from the cache
        cnx = sqlite3.connect(DB_PATH)
        cursor = cnx.cursor()
        if query.symbol:
            cursor.execute(
                "SELECT * FROM short_interest where symbolCode = ?", (query.symbol,)
            )
        else:
            cursor.execute("SELECT * FROM short_interest")
        result = cursor.fetchall()

        titles = [
            "symbolCode",
            "issueName",
            "marketClassCode",
            "currentShortPositionQuantity",
            "previousShortPositionQuantity",
            "averageDailyVolumeQuantity",
            "daysToCoverQuantity",
            "changePercent",
            "changePreviousNumber",
            "settlementDate",
        ]
        return [
            {title: value for title, value in zip(titles, list(row)[1:])}
            for row in result
        ]

    @staticmethod
    def transform_data(
        query: FinraShortInterestQueryParams, data: List[Dict], **kwargs: Any
    ) -> List[FinraShortInterestData]:
        """Transform the data."""
        return [FinraShortInterestData.model_validate(d) for d in data]



"""FINRA provider module."""
from openbb_core.provider.abstract.provider import Provider
from openbb_finra.models.equity_short_interest import FinraShortInterestFetcher
from openbb_finra.models.otc_aggregate import FinraOTCAggregateFetcher

finra_provider = Provider(
    name="finra",
    website="https://finra.org",
    description="Financial Industry Regulatory Authority.",
    credentials=None,
    fetcher_dict={
        "OTCAggregate": FinraOTCAggregateFetcher,
        "EquityShortInterest": FinraShortInterestFetcher,
    },
)
